/*************************************************************************/
%                                                                                                 
% Main.pl (Automatische Stundenplangenerierung)                                                   
%                                                     
% Dieses File enthlt die Prdikate zum Start der Stundenplangenerierung.
%                                                                                                 
/*************************************************************************/

:- use_module(library(fd)).

:- dynamic vorlesungAm/2.
:- dynamic vorlesungNichtUm/2.
:- dynamic vorlesungNichtAm/2.
:- dynamic raum/3.
:- dynamic vorlesung/4.
:- dynamic dummyVorlesung/1.
:- dynamic allBlocksSameDay/1.
:- dynamic mittagszeit/1.
:- dynamic outputFileName/1.

  
/*************************************************************************/
%
% startGeneration/1(+DB)
%
% Startet die Generierung des Stundenplanes mit der Datenbasis DB 
% (wird ber consult geladen !).                                             
% Als Output-Stream fr die Meldungen des Systems wird StdOut verwendet.
%
% DB : Filename der Datenbasis
%
%*************************************************************************/
startGeneration(DB) :-
   get_stream(output,OutputStream),
   startGeneration(DB,OutputStream).

/*************************************************************************/
%
% startGeneration/2(+DB,+STREAM)
%
% Wie startGeneration/1 nur kann hier noch der Ausgabestream fr Meldungen
% angegeben werden
%
% DB     : Filename der Datenbasis
% STREAM : Stream-Handle fr den Stream auf dem Systemmeldungen ausgegeben
%          werden.
%
%*************************************************************************/
startGeneration(DB,STREAM) :-
   clearDB,
   asserta(outputStream_(STREAM)),
   message(['Starte Stundenplanerstellung ...']),
   compile(DB),
   compile('checkDB'),                 % Prdikate zum Prfen der Datenbasis
   compile('constraints'),             % Allgemeine Constraints
   compile('internalStructures'),      % Aufbau von internen Strukturen
   compile('HTMLOutput'),              % Prdikate zur HTML-Ausgabe des Ergebnisses
   compile('specificConstraints'),     % Prdikate fr die Verarbeitung der
                                       %  spezifischen Constraints der DB
   compile('search'),                  % Suchprdikate
   compile('DummyLectures'),           % Prdikate fr Scheinvorlesungen
   message(['Lese Datenbasis ',DB]),
   ( checkDB -> 
     (
       message(['Bereite interne Datenstrukturen vor ...']),
       !,prepareInternals,
       message(['Stundenplanerstellung ...']),
       !,startTimetabling
     );
     message(['Datenbasis nicht Fehlerfrei'])
   ).


/*************************************************************************/
%
% prepareInternals/0
%
% Bereitet Aufbauend auf der eingelesenen Datenbasis die internen Strukturen
% vor.
%
% (1) Indiziert und zhlt die in der Datenbasis (DB) definierten
%     Tage, Zeiten und Rume
% (2) Veranlasst die Erstellung von Scheinvorlesungen fr die Mittagspause 
%     eines jeden Tages fr alle Gruppen.
%
% (3) Macht die Ergebnisliste der Prozedur preparedCourseList/1 unter dem Fakt
%     preparedCourses/1 in der Klauselmenge verfgbar.
%
% (4) Macht die Ergebnisliste der Prozedur blockNumberList/1 unter dem Fakt
%     blockNumners/1 in der Klauselmenge verfgbar.
%
%
%*************************************************************************/
prepareInternals :-
	asserta(dayNumbers(dummy)),
	retract_all(dayNumbers(_)),
	dayNumberList(DayNumbers,DayBlockFactors),
	asserta(dayNumbers(DayNumbers)),

        asserta(noOfDays(dummy)),
	retract_all(noOfDays(_)),
	length(DayNumbers,NoOfDays),
	asserta(noOfDays(NoOfDays)),

        asserta(timeNumbers(dummy)),
	retract_all(timeNumbers(_)),
	timeNumberList(TimeNumbers,TimeBlockFactors),
	asserta(timeNumbers(TimeNumbers)),

	message(['Erstelle Scheinvorlesungen ...']),
        !,prepareDummyLectures,  
	
	asserta(roomNumbers(dummy)),
	retract_all(roomNumbers(_)),
	roomNumberList(RoomNumbers,RoomBlockFactors),
	asserta(roomNumbers(RoomNumbers)),

        asserta(noOfRooms(dummy)),
	retract_all(noOfRooms(_)),
	length(RoomNumbers,NoOfRooms),
	asserta(noOfRooms(NoOfRooms)),
	
	asserta(noOfDummyRooms(dummy)),
	retract_all(noOfDummyRooms(_)),
	findall(R,dummyRaum(R),Rs),
	length(Rs,NoOfDummyRooms),
	asserta(noOfDummyRooms(NoOfDummyRooms)),

        /* Da retract_all Exceptions generiert wenn keine zu entfernenden Klauseln in der  */
        /* Datenbasis vorhanden sind, wird jeweils vor retract_all ein entspr. asserta     */
        /* eingefgt.                                                                      */

	asserta(preparedCourses(dummy)), 
	retract_all(preparedCourses(_)),
	preparedCourseList(PreparedCourses),
	asserta(preparedCourses(PreparedCourses)),

	/*  course\4(Name,Groups,Resources,Blocks)        */
	/*  Resources enthlt die size\1 Resource         */
	/*  Liste ist nach Kursgrsse absteigend sortiert */

        asserta(blockNumbers(dummy)),
	retract_all(blockNumbers(_)),
	blockNumberList(PreparedCourses,BlockNumbers),
	asserta(blockNumbers(BlockNumbers)),       
 
	/*  block\3(BlockNr,Vorlesungsname,BlockStruktur)*  */
	/*  Liste wurde aus PreparedCourses\1 erstellt und ist also auch  */
	/*  mit absteigender Kursgrsse sortiert */
  
        asserta(noOfCourses(dummy)),
	retract_all(noOfCourses(_)),
	length(PreparedCourses,NoOfCourses),
	asserta(noOfCourses(NoOfCourses)),
	
	asserta(noOfTimesToSchedule(dummy)),
	retract_all(noOfTimesToSchedule(_)),
	timesToSchedule(T),               
        /* Die Gesamtzahl aller Doppelstunden die zu terminieren sind */
	
        asserta(noOfTimesToSchedule(T)),

        asserta(commonGroups(dummy)),
	retract_all(commonGroups(_)),!,
	notTheSameTime(PreparedCourses,PairList),
	asserta(commonGroups(PairList)),  
        
        /* Diese Liste enthlt alle Paare von Veranstaltungen, die */
        /* jeweils nicht zu einem Zeitpunkt stattfinden drfen.    */

	append(DayBlockFactors,TimeBlockFactors,BlockFactors0),
	append(BlockFactors0,RoomBlockFactors,BlockFactors),
	asserta(blockFactors(dummy)),
	retract_all(blockFactors(_)),
	asserta(blockFactors(BlockFactors)),

        asserta(noOfTimes(dummy)),
	retract_all(noOfTimes(_)),
	length(TimeNumbers,NoOfTimes),
	asserta(noOfTimes(NoOfTimes)),

        asserta(noOfDaysAndTimes(dummy)),
  	retract_all(noOfDaysAndTimes(_)),
	NoOfDaysAndTimes is NoOfDays * NoOfTimes,
	asserta(noOfDaysAndTimes(NoOfDaysAndTimes)).



/*************************************************************************/
%
% startTimetabling/0
%
% Startet den eigentlichen Generierungsprozess.
% - Erstellt die Constraintvariable
% - Baut ber den Variablen die allgemeinen und spezifischen Constrints auf
% - Startet dann den Suchprozess
% - Gibt das Ergebnis aus
% Wird von startGeneration/1/2 aufgerufen nachdem die internen Datenstrukturen
% vorbereitet wurden. 
%
%*************************************************************************/
startTimetabling :-
        message(['Erstelle Constraintvariable ...']),
	blockNumbers(Blocks),   
	mkVariables(Blocks,Vars),

	extractNonDummies(Vars,NonDummies,_),
 	mkXVariables(NonDummies,XVars), 

	preparedCourses(Courses), 

        message(['---- Constraints ber den Variablen aufbauen ----']),
	message(['Einschrnkungen bzgl. der Rume ...']),
	!,roomConstraints(Courses,Vars),
        
	message(['Einschrnkungen bzgl. der Zeit ...']),
	!,timeConstraints(Vars),

	message([' ... Einschrnkungen bzgl. der Blockstruktur ']),
	!,blockConstraints(Vars), 
				  
	alldifferent(XVars),       

        % Verknpft alle mit mkXVariables aufgebauten Einschrnkungen
	% d.h. in einem Raum kann nur max. eine Vorlesung zur Zeit
	% stattfinden.

	message(['Wende spezifische Constraints an ...']),
	!,applySpecificConstraints(Vars), 

	message(['Suche im Zustandsraum ...']),
	!,search(Vars),
        message(['Ergebnis : ']),
	asserta(timetable(dummy)), 
	retract_all(timetable(_)),
	asserta(timetable(Vars)),

   	out(Vars),
	message(['Erstelle HTML-Output ']),!,
	getOutSkeleton(html,OutList,Holes,'Stundenplan',Vars),
	message(['HTML-Output fertiggestellt ']),!,
	fillInHoles(Vars,Holes),
	message(['HTML Tabelle ausgefllt ']),
	(outputFileName(ResultFileName) -> true ; ResultFileName='result.html'),
	open(ResultFileName,write,Stream),
	writeOutListToStream(Stream,OutList),
	close(Stream),
	message(['---ENDE---']),

	true.
	
/*************************************************************************/
%
% fillInHoles/2(+DTRList,+HolesList)
%
% Fllt die mit der Prozedur getOutSkeleton/5 erstellten freien Variablen
% aus, also den auszugebenden Stundenplan.
% Fr jede dtr/4 Struktur aus der DTRList werden die entsprechenden Variablen
% aus der HolesList ausgefllt.
%
% DTRList : Liste von dtr/4 Strukturen, nach dem Suchprozess.
%           (Diese Liste ist das Ergebnis der Stundenplanung)
% HolesList : Liste von hole/2 Strukturen wie sie von der Prozedur 
%             getOutSkeleton geliefert wird.
%
%*************************************************************************/
fillInHoles([],_).
fillInHoles([dtr(D,Ti,R,BID)|T],Holes) :-
    blockNumbers(Blocks), member(block(BID,VName,_),Blocks),
    not(dummyVorlesung(VName)),
    member(hole(dtr(D,Ti,R),Value),Holes),!,
    nameForBID(BID,Value),
    member(hole(grp(BID),GruppenListe),Holes),
    vorlesung(Value,Gruppen,_,_),
    prepareGroupList(Gruppen,GruppenListe),
    fillInHoles(T,Holes).
fillInHoles([dtr(_,_,_,_)|T],Holes) :-
    fillInHoles(T,Holes).

%Bereitet die bergebene Liste von Gruppen fr die Ausgabe im Ausgabeformat
%des Stundenplanes aus, indem Dummy-Gruppen entfernt, und zwischen den anderen
%Gruppen Spaces eingefgt werden.

prepareGroupList([],[]).
prepareGroupList([H|T],Res):-
    dummyGroup(H),!,
    prepareGroupList(T,Res).
prepareGroupList([H|T],Res) :-
    prepareGroupList(T,Res0),
    append([' ',H],Res0,Res).


/*-------------------------------------------------------------------------------------------*/
/*                                                                                           */
/* Prdikate zur Ausgabe von Messages sowie des formatierten Ergebnisses                     */
/*                                                                                           */
/*-------------------------------------------------------------------------------------------*/

out(Vars) :-
    outcourses(Vars,1),
    asserta(nichtZuweisbar(dummy)),
    findall(BID,(nichtZuweisbar(BID),not(=(BID,dummy))),BIDS),
    procNichtZuweisbar(BIDS),
    asserta(nichtZuweisbar(dummy)),
    retract_all(nichtZuweisbar(_)).

procNichtZuweisbar([]).
procNichtZuweisbar([BID|T]) :-
    nameForBID(BID,Name),
    !,message([' Vorlesung nicht zuweisbar -> ',Name]),
    procNichtZuweisbar(T).

outcourses([],_).
outcourses([dtr(Dn,Tn,Rn,BID)|Vars],I) :-
        dayNumber(D,Dn),
	timeNumber(T,Tn),
	roomNumber(R,Rn),
	nameForBID(BID,VName),
	write('Vorlesung BlockNr : '),write(VName),write(' '),write(BID),write(' am '),
	write(D),write(' um '),
	write(T),write(' im Raum '),
	write(R),nl,
	NEXT is I + 1,
	outcourses(Vars,NEXT).
outcourses([dtr(_,_,_,BID)|Vars],I) :-
    asserta(nichtZuweisbar(BID)),
    NEXT is I + 1,
    outcourses(Vars,NEXT).
    
/************************************************************************************************/
/* Gibt die Bestandteile der Liste mit Hilfe des write-Prdikates auf dem Stream aus der in der */
/* Datenbasis als outputStream_(Out) verzeichnet ist.                                           */
/************************************************************************************************/
message([]) :-
	outputStream_(Out),
	nl(Out).

message([H|R]) :-
	outputStream_(Out),
	write(Out,H),
	message(R).


/*************************************************************************/
%
% clearDB/0
%
% Entfernt die Klauseln
%  raum/3,dummyRaum/1,getNoMiddayTimes/1,vorlesung/3,dummyVorlesung/1
%  alleBloeckeAmSelbenTag/1,vorlesungNichtUm/1
% aus der Datenbasis. Wird von startGeneration/1/2 als erstes aufgerufen
% um dynamische Klauseln aus eventuellen vorausgegangenen Lufen aus der
% Datenbasis zu entfernen.
%
%*************************************************************************/
clearDB :-
    asserta(raum(dummy,dummy,dummy)),
    asserta(dummyRaum(dummy)), 
    asserta(getNoMiddayTimes(dummy)),
    asserta(vorlesung(dummy,dummy,dummy,dummy)),
    asserta(dummyVorlesung(dummy)),
    asserta(alleBloeckeAmSelbenTag(dummy)),
    asserta(vorlesungNichtUm(dummy)),
    retract_all(raum(_,_,_)),
    retract_all(dummyRaum(_)),
    retract_all(getNoMiddayTimes(_)),
    retract_all(vorlesung(_,_,_,_)),
    retract_all(dummyVorlesung(_)),
    retract_all(alleBloeckeAmSelbenTag(_)),
    retract_all(vorlesungNichtUm(_)).
    
/*************************************************************************/
%
% extractNonDummies/3(+DTRList,?NonDummies,?Dummies)
%
% Unterteilt die dtr/4 Liste DTRList in dtr/4 Strukturen die zu regulren
% Vorlesungen gehren und solche die zu Vorlesungen gehren, die als 
% dummyVorlesung/1 definiert sind.
%
% +DTRList Die dtr/4 Liste die von mkVariables erstellt wurde und
%          im System verwendet wird.
% ?NonDummies Liste von DTR/4 Strukturen die zu regulren Vorlesungen gehren
% ?Dummies    Liste von dtr/4 Strukturen die zu Vorlesungen gehren,
%             die als dummyVorlesung/1 definiert wurden.
%
%*************************************************************************/
extractNonDummies([],[],[]).
extractNonDummies([dtr(Day,Time,Room,BID)|Rest],NonDummies,Dummies) :-
    blockNumbers(Blocks), member(block(BID,VName,_),Blocks),
    dummyVorlesung(VName),!,    
    extractNonDummies(Rest,NonDummies,Dummies0),
    append([dtr(Day,Time,Room,BID)],Dummies0,Dummies).
extractNonDummies([DTR|Rest],NonDummies,Dummies) :-
    extractNonDummies(Rest,NonDummies0,Dummies),
    append([DTR],NonDummies0,NonDummies).

    







