%
% search.pl (Automatische Stundenplangenerierung)
% 
% Prdikate fr die Suche im Suchraum des Problems 


/*************************************************************************/
%
% search/1(+Vars)
%
% Sucht mit Hilfe einer direkten Heuristik und lokaler BranchAnBound-Suche 
% nach einer Lsung, d.h. nach einem durchfhrbaren Stundenplan.
% Die textuellen Ausgaben des Suchprozesses werden in das File "search.txt"
% geschrieben. Wird von startTimetabling/0 aufgerufen.
%
% +Vars Die dtr/4-Constraintvariablenliste, mit Anfangs noch nicht festgelegten
%       Variablenwerten. Nachdem die Suche abgeschlossen ist, sind aller Variablen
%       deren zugehrige Blcke verplant werden konnten (i.a. alle) an Werte 
%       gebunden, und die Liste stellt eine Lsung des Problems, also einen 
%       Stundenplan dar.
%
%
/*************************************************************************/
search(Vars) :-
    courseSelectionList(BIDs,Vars),
    searchRec(Vars,BIDs).

/******************* SUCHE ***********************************************/    
/*************************************************************************/
%
% searchRec/2(+Vars,+BIDList)
%
% Wird von search/1 benutzt, um die eigentliche Suche durchzufhren, und
% benutzt dazu die Liste von BlockID's BIDList um die Reihenfolge der 
% Verplanung von Vorlesungen zu erhalten (BIDList wurde von der Proc. 
% CourseSelectionList/2 erstellt). Siehe auch search/1
%
% +Vars Die dtr/4-Constraintvariablenliste, mit Anfangs noch nicht festgelegten
%       Variablenwerten. Nachdem die Suche abgeschlossen ist, sind aller Variablen
%       deren zugehrige Blcke verplant werden konnten (i.a. alle) an Werte 
%       gebunden, und die Liste stellt eine Lsung des Problems, also einen 
%       Stundenplan dar.
%
% +BIDList Liste von BlockIDs die festlegt welche Blcke in welcher Reihenfolge
%          verplant werden (Die gesamte Liste wurd von der proc. 
%          courseSelectionList/2 erstellt).
/*************************************************************************/
searchRec(_,[]).
searchRec(Vars,[BID|Rest]) :-	
    getDTRForBID(dtr(Day,Time,Room),BID,Vars), 
    helpOut0(BID),!,
    minimize(localSearch(Day,Time,Room,Cost,Vars,BID),Cost), 
    out_(Day,Time,Room,BID), 
    searchRec(Vars,Rest).
searchRec(Vars,[BID|Rest]) :-
    nameForBID(BID,Name),
    message(['FEHLER -> Konnte Vorlesung ',Name,' BID ',BID,' nicht zuweisen']),
    searchRec(Vars,Rest).

getDTRForBID(dtr(D,T,R),BID,Vars) :-
    member(dtr(D,T,R,BID),Vars).

%
% Sucht eine mgliche Belegung der Constraint Variablen +D (Tag) +T (Zeit) +R (Raum) des 
% Blockes mit der BlockId BID. Die Kosten dieser Belegung werden ber -Cost zurckgegeben
% +Vars ist die dtr/4-Constraintvariablenliste
%
localSearch(D,T,R,Cost,Vars,BID) :-
    message(['LocalSearch ',BID]),!,
    indomain(D),
    indomain(T),
    indomain(R),
    calcLocalCost(D,R,Cost,Vars,BID).
    
%
% Berechnet die Kosten -Cost wenn der Block mit der BlockId +BID auf Tag +D und Raum +R gelegt wird.
% +Vars ist die dtr/4 Constraintvariablenliste. Dabei werden die Blockfaktoren der einzelnen Werte
% der Wertebereiche der Constraintvariablen bercksichtigt.
%
calcLocalCost(D,R,Cost,Vars,BID) :-
    not(mittagspause(BID)),!,
    findall(Factor,(member(dtr(_,Time,_,BID),Vars),nameID(TName,Time,time),blockFactor(TName,Factor)),TFactors),
    sumFactors(TFactors,TimeFactor),
    nameID(DayName,D,day),nameID(RoomName,R,room),
    blockFactor(DayName,DayFactor),blockFactor(RoomName,RoomFactor),
    Cost is DayFactor+TimeFactor+RoomFactor.
calcLocalCost(_,_,Cost,Vars,BID) :-
    firstOccurrence(dtr(_,_,_,BID),Vars,dtr(_,T,_,_)),
    nameID(Time,T,time),
    valueForMiddayTime(Time,Cost).


% Sucess wenn BID eine Mittagspausen Dummyvorlesung ist.
%
mittagspause(BID):-
    blockNumbers(Blocks),
    member(block(BID,['Mittag',_,_],_),Blocks).

sumFactors([],0).
sumFactors([H|T],Res):-
    sumFactors(T,Res0),
    Res is Res0+H.

out_(Day,Time,Room,BID) :-    
    dayNumber(DayNam,Day),timeNumber(TimeNam,Time),roomNumber(RaumNam,Room),
    nameForBID(BID,VorlName),message([' Vorlesung :',VorlName,' auf ',DayNam,' ',TimeNam,' in ',RaumNam]).
out_(_,_,_,_) :- fail.

helpOut0(BID) :-
    nameForBID(BID,VName),
    message(['-----------------------------------------------------------------------']),
    message([' BID ',BID,' Vorlesung ',VName]).
helpOut0(_) :- fail.

%
% Liefert den BlockFaktor der fr den bergebenen Wert eines Wertebereiches (Tage,Zeiten oder Rume)
% definiert ist.
%
blockFactor(ValueName,Factor) :-
    blockFactors(BF),
    member(blockFactor(ValueName,Factor),BF).


/******************************************************/
/* courseSelectionList/1(-BIDS)                       */
/* Liefert alle zu terminierenden BID's in der        */
/* Reihenfolge in der sie terminiert werden sollen.   */
/* Die BID's sind in der Liste mit absteigender       */
/* Priorit (bezogen auf die Terminierungsreihenfolge)*/
/* angelegt.                                          */
/******************************************************/ 
courseSelectionList(BIDS,Vars) :-
    extractNonDummies(Vars,_,Dummies),
    findall(DummyBID,member(dtr(_,_,_,DummyBID),Dummies),DummyBIDs),
    /*------------------------------------------------*/
    /* 1. Schritt  Klassifizierung nach Blockgroesse  */
    /*------------------------------------------------*/
    blockNumbers(B),    
    findall(BlockSize-Name,
            (  
	       vorlesung(Name,_,_,_),
	       not(dummyVorlesung(Name)),
	       member(block(ABid,Name,_),B),	       
	       blockSize(ABid,BlockSize)
	    ),
	    Courses),
    keysort(Courses,Courses0),
    findall(BlockSize-BID,
            (member(BlockSize-Name,Courses0),aBIDForCourse(Name,BID)),
	    BIDsorted    
           ),
    classify(BIDsorted,1,Classes0),
    reverse(Classes0,Classes),
    
    courseSelectionList_2(BID_S,Classes,Vars),
    append(BID_S,DummyBIDs,BIDS).   /*Mittagspausen ganz zum Schluss schedulen */
    
aBIDForCourse(Name,BID) :-
    blockNumbers(Blocks),!,
    member(block(BID,Name,_),Blocks).

courseSelectionList_2([],[],_).
courseSelectionList_2(BIDS,[Class|Classes],Vars) :-  /* Jede Klasse enthlt BIDs */
    /*---------------------------------------------------------*/
    /* 2. Schritt Unterklassifizierung nach Gruppen            */
    /*---------------------------------------------------------*/
    blockNumbers(B),
    findall(NumOfGroups-Bid,
              (
	        member(Bid,Class),
		member(block(Bid,VName,_),B),
		vorlesung(VName,Groups,_,_),
		numberOfGroups(Groups,NumOfGroups)
	      ),BIDList),
    keysort(BIDList,BIDList0),
    
    classify(BIDList0,1,BlockClasses0),
    reverse(BlockClasses0,BlockClasses),

    courseSelectionList_3(BIDS_0,BlockClasses,Vars),
    
    courseSelectionList_2(BIDS_1,Classes,Vars),
    append(BIDS_0,BIDS_1,BIDS).
    
courseSelectionList_3([],[],_).
courseSelectionList_3(BIDS,[Class|Classes],Vars) :-
    /*--------------------------------------------------------------*/
    /* 3. Schritt Unterklassifizierung nach Resourcenanforderungen  */
    /*--------------------------------------------------------------*/
    findall(Factor-Bid,(member(Bid,Class),resourceFactor(Bid,Vars,Factor)),BIDList),
    keysort(BIDList,BIDList0),
    
    classify(BIDList0,1,ResourceClasses),
    courseSelectionList_4(BIDS_0,ResourceClasses),
    courseSelectionList_3(BIDS_1,Classes,Vars),
    
    append(BIDS_0,BIDS_1,BIDS).


courseSelectionList_4([],[]).    
courseSelectionList_4(BIDS,[Class|Classes]) :-
    courseSelectionList_4(BIDS_0,Classes),
    append(Class,BIDS_0,BIDS).
    


%
% Liefert die Anzahl -NumOfGroups der Gruppen in der Liste von Gruppen +Groups.
%
numberOfGroups(Groups,NumOfGroups) :-  
    removeDoubles(Groups,Groups0),
    length(Groups0,NumOfGroups).

%
% Liefert die Grsse -SIZE in Vorlesungseinheiten des Blocks mit der BlockId +BID.
%
blockSize(BID,SIZE) :-
    blockNumbers(Blocks),
    member(block(BID,_,Structure),Blocks),
    blockStructureSize(Structure,SIZE).

%
% Berechnet einen Faktor -Factor der Resourcenanforderungen des Blocks mit der BlockId +BID.
% +Vars ist die dtr/4-Constraintvariablenliste. Der Faktor entspricht der momentanen Kardinalitt des
% Wertebereiches der Raum-Constraintvariable des Blockes. 
%
resourceFactor(BID,Vars,Factor) :-
    member(dtr(_,_,Room,BID),Vars),!,
    dvar_domain(Room,RoomDomain),
    dom_size(RoomDomain,Factor).

%
% Der Name ?Name gehrt zum Index ?ID der abhngig vom 3. Paramerter (time oder day oder room) 
% ein Zeit,Tag oder Raumindex ist. Wird verwendet um Indizies von Tagen,Zeiten und Rumen auf ihre
% Namen und umgekehrt abzubilden.
%
nameID(Name,ID,time) :-
    timeNumber(Name,ID).
nameID(Name,ID,day) :-
    dayNumber(Name,ID).
nameID(Name,ID,room) :-
    roomNumber(Name,ID).

/***********************************************************/
/* classify/3(+KeyValueList,+ClassSize,-Classes)           */
/*                                                         */
/* Nimmt eine Liste von Schlssel Wert Paaren der Form     */
/* Schlssel-Wert entgegen, wobei der Schlssel ein        */
/* Integer ist und die Listenentrge nach Schlsseln       */
/* aufsteigend sortiert sein mssen.                       */
/* -Classes ist die Liste von Klassen (enthalten die Werte */
/* aus den Schlssel-Wert Paaren) , die sich nach          */
/* der +KeyValueList nach Keys in Intervalle der Grsse    */
/* +ClassSize ergeben haben.                               */
/***********************************************************/
classify(KeyValueList,ClassSize,Classes) :-
    KeyValueList=[Key-Value|KeyValues],!,
    NextThreshold is Key+ClassSize,
    classify(KeyValues,ClassSize,[],Classes,NextThreshold,[Value]).
classify([],_,[]).
    
classify([],_,AccClasses,ResClasses,_,ActClass) :- append(AccClasses,[ActClass],ResClasses).
classify([Key-Value|KeyValues],ClassSize,AccClasses,ResClasses,Threshold,ActClass) :-
    Key<Threshold,
    append(ActClass,[Value],ThisClass),
    classify(KeyValues,ClassSize,AccClasses,ResClasses,Threshold,ThisClass).
classify([Key-Value|KeyValues],ClassSize,AccClasses,Classes,Threshold,ActClass) :-
    Key>=Threshold,
    append(AccClasses,[ActClass],AccClasses0),
    NextThreshold is Threshold+ClassSize,
    classify(KeyValues,ClassSize,AccClasses0,Classes,NextThreshold,[Value]).    



valueForMiddayTime('Mittagspause',0).
valueForMiddayTime('11:00',1).
valueForMiddayTime('13:45',1).

