%
% checkDB.pl (Stundenplangenerierung)
%
% Prdikate zum Plausibilitscheck der Datenbasis.                    
% Geht davon aus, da die DB schon compiliert ist !!                  


%
% Ruft checkDB/1 mit dem Stream outputStream_/1 auf.
%
checkDB :- 
    outputStream_(OS),
    !,checkDB(OS).

/*****************************************************************************************/
%
% checkDB/1(+ErrMsgStream)
%
% berprft die Datenbasis, und schreibt textuelle Ausgaben auf den bergebenen Stream
% +ErrMsgStream.
% Success wenn :
%
%   - Keine doppelten Raumnamen vorhanden
%   - Keine doppelten Vorlesungsnamen vorhanden.
%   - Alle verwendeten Resourcen/Gruppen definiert wurden.
%   - Nur gltige Blockdefinitionen verwendet wurden. (Par. 4 von vorlesung/4)
%   - Die Resourcenliste der Rume als erste Resource die Struktur groesse/1 enthlt.
%
/*****************************************************************************************/
checkDB(ErrMsgStream) :-

    outputStream_(OldStream),
    retract_all(outputStream_(_)),
    asserta(outputStream_(ErrMsgStream)),
 
    allUsedGroups(AllUsedGroups),       allUsedResources(AllUsedResources),
    allUsedBlockings(AllUsedBlockings), allUsedLectureNames(AllUsedLectureNames),
    allUsedRoomNames(AllUsedRoomNames), 

    doubles(AllUsedRoomNames,DoubleRooms),
    !, check(DoubleRooms,['OK - Keine Doppelten Raumnamen'],['FEHLER - Doppelte Raumnamen ',DoubleRooms],Errors),

    doubles(AllUsedLectureNames,Doubles),
    !,check(Doubles,['OK - Keine Doppelten Vorlesungsnamen'],['FEHLER - Doppelte Vorlesungsnamen ',Doubles],Errors),
  
    allDefinedGroups(ADG),
    notContained(AllUsedGroups,ADG,MissingGroups),
    !,check(MissingGroups,['OK - Alle verwendeten Gruppen definiert'],
                        ['FEHLER - Nicht definierte Gruppen ',MissingGroups],Errors),

    allDefinedResources(ADR),
    notContained(AllUsedResources,ADR,MissingResources),
    !,check(MissingResources,['OK - Alle verwendeten Resourcen vorhanden (Kombinationen nicht geprft !)'],
                           ['FEHLER - Fehlende Resourcen ',MissingResources],Errors),
    
    allAvailableBlockings(AAB),
    notContained(AllUsedBlockings,AAB,UnknownBlockings),
    !,check(UnknownBlockings,['OK - Alle verwendeten Blockungen bekannt'],
                           ['FEHLER - Unbekannte Blockungen ',UnknownBlockings],Errors),
    
    !,checkAllRooms,

    retract_all(outputStream_(_)),
    asserta(outputStream_(OldStream)),

    var(Errors).
    
allUsedGroups(AUG) :-
    findall(Groups,vorlesung(_,Groups,_,_),GroupsList),
    findall(L1,dozentNichtAm(L1,_),List1),
    findall(L2,dozentNichtUm(L2,_),List2),
    findall(L3,dozentNichtAmUm(L3,_,_),List3),
    List4=[GroupsList,List1,List2,List3],
    ensureIsList(List4,List5),
    removeDoubles(List5,AUG).

allUsedResources(AUR) :-
    findall(R,vorlesung(_,_,R,_),RList),
    ensureIsList(RList,ResList),
    removeDoubles(ResList,AUR).
    
allUsedBlockings(AUB) :-
    findall(B,vorlesung(_,_,_,B),BList),
    ensureIsList(BList,Blockings),
    removeDoubles(Blockings,AUB).
    
allUsedLectureNames(AUN) :-
    findall(Name,vorlesung(Name,_,_,_),AUN).
    
allUsedRoomNames(ARN) :-
    findall(Name,raum(Name,_,_),ARN).    

allDefinedGroups(ADG) :-
    findall(Group,group(Group,_),Groups),
    findall(SubGroup,subgroup(SubGroup,_,_,_),SubGroups),
    findall(Lecturer,dozent(Lecturer),Lecturers),
    append(Groups,Lecturers,ADG0),
    append(ADG0,SubGroups,ADG).

allDefinedResources(ADR) :-
    findall(Resources,raum(_,Resources,_),ResList0),
    ensureIsList(ResList0,ResList),
    removeAll(groesse(_),ResList,List),
    removeDoubles(List,ADR).    
    
allAvailableBlockings(AAB) :-
    AAB=[single(_),singleNoCluster(_),
         block2(_),block2NoCluster(_),
	 block3(_),block3NoCluster(_)].

doubles(List,Doubles) :-
    collectDoubles(List,Collected),
    removeDoubles(Collected,Doubles).

collectDoubles([],[]).
collectDoubles([H|T],[H|R]) :-
    member(H,T),!,
    collectDoubles(T,R).
collectDoubles([_|T],Doubles) :-
    collectDoubles(T,Doubles).
    
notContained(A,B,Missing) :-
    notContained_(A,B,Missing0),
    removeDoubles(Missing0,Missing).

notContained_([],_,[]).    
notContained_([freizeit|T],B,Missing) :-
    notContained_(T,B,Missing).
notContained_([oder(L)|T],B,Missing) :-
    checkOrStructure(L,B,Missing0),
    notContained_(T,B,Missing1),
    append(Missing0,Missing1,Missing).
notContained_([H|T],B,[H|R]) :-
    not(member(H,B)),!,
    notContained(T,B,R).
notContained_([_|T],B,Missing) :-
    notContained_(T,B,Missing).    
    
checkOrStructure(L,B,Missing) :-
    L=[_|_],!,
    notContained(L,B,Missing).
checkOrStructure(L,_,[]) :- 
    message(['oder/1 Struktur oder(',L,') Parameter muss eine Liste von Resourcen sein']).
    
checkAllRooms :-
    findall(raum(Name,Res,BlockFact),raum(Name,Res,BlockFact),Rooms),
    checkAllRooms(Rooms).
checkAllRooms([]).
checkAllRooms([raum(Name,Res,BlockFact)|T]) :-
    checkResources(Res,Name),
    checkBlockFactor(BlockFact,Name),
    checkAllRooms(T).
    
checkResources([groesse(_)|_],_).
checkRsources(_,Name) :- 
    message(['Resourcendefinition von Raum ',Name,' ohne Resource groesse/1 ']).

checkBlockFactor(BF,_) :- BF >= 0.
checkBlockFactor(_,Name) :-
    message(['Blockfaktor von Raum ',Name,' ist kein positiver Integerwert']).
    
removeAll(_,[],[]).
removeAll(E,[E|T],Result) :-    
    removeAll(E,T,Result).
removeAll(E,[H|T],[H|Res0]) :-	
    removeAll(E,T,Res0).
    
check(List,MsgLength0,_,_) :- 
    length(List,0),!,
    message(MsgLength0).
check(_,_,OtherMsg,true) :-
    message(OtherMsg).











